/*
	Author: James Sumners (james dot sumners AT gmail dot com)
	URL: http://james.roomfullofmirrors.com/
	License: Creative Commons 3.0 (NCSA)
		http://creativecommons.org/licenses/by-nc-sa/3.0/us/
	Attribution: Provide a name credit and link to URL above.
*/

#include <stdio.h>
#include <string.h>
#include "tiffio.h"
#include "globals.h"
#include "debug.h"

void write_matrix(img_f matrix[][width], char file[]) {
	int i, j;
	FILE *file_out;
	file_out = fopen(file, "w");
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			fprintf(file_out, "%06.3f ", matrix[i][j]);
		}
		fprintf(file_out, "\n");
	}
	fclose(file_out);
}

void write_matrix_mapmat(img_f matrix[][width], char file[], char *mapmat) {
	int i, j;
	FILE *file_out;
	file_out = fopen(file, "w");
	
	if (strcmp(mapmat,"maple")) {
		for (i = 0; i < height; i++) {
			for (j = 0; j < width; j++) {
				if (j != (width-1)) {
					fprintf(file_out, "%i\t", (int)matrix[i][j]);
				} else {
					fprintf(file_out, "%i", (int)matrix[i][j]);
				}
			}
			if (i != (height-1)) {
				fprintf(file_out, "\n");
			}
		}
	} else {
		for (i = 0; i < height; i++) {
			for (j = 0; j < width; j++) {
				if ( j != width-1 ) {
					fprintf(file_out, "%i ", (int)matrix[i][j]);
				} else if ( j == width-1 ) {
					if (i != height-1) {
						fprintf(file_out, "%i\n", (int)matrix[i][j]);
					} else {
						fprintf(file_out, "%i", (int)matrix[i][j]);
					}
				}
			}
		}
	}
	
	fclose(file_out);
}

void tiff_to_matrix(img_c **buffer, img_f matrix[][width]) {
	int i, j, k;
	
	/* Copying the image data into a "matrix" so it is easier
	   to work with. */
	i = k = 0;
	while (k < width*height) {
		for (j = 0; j < width; j++) {
			matrix[i][j] = (float)(*buffer)[k];
			k++;
		}
		i++;
	}
}

void matrix_to_tiff(img_f matrix[][width], img_c **buffer) {
	int i, j, k;
	
	/* Copy the data back to the image buffer. */
	i = k = 0;
	while (k < width*height) {
		for (j = 0; j < width; j++) {
			(*buffer)[k] = floorf(matrix[i][j]);
			k++;
		}
		i++;
	}
}

void count_zeros(img_f matrix[][width], int version) {
	int i, j;
	int m = 0, n = 0;
	
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			if ((int)matrix[i][j] == 0) {
				m++;
			} else {
				n++;
			}
		}
	}
	
	if (version == 0) {
		zeros.original_zeros = m;
		zeros.original_nonzeros = n;
	} else if (version == 1) {
		zeros.compressed_zeros = m;
		zeros.compressed_nonzeros = n;
	} else {
		zeros.restored_zeros = m;
		zeros.restored_nonzeros = n;
	}
}

float calculate_mse(img_f matrix[][width], img_f original[][width]) {
	img_f C[height][width];
	img_f V[width];
	float sum = 0;
	int i, j;
	
	/* Initialize V. */
	for (j = 0; j < width; j++) {
		V[j] = 0;
	}
	
	/* Difference matrices, square C, and sum C columns. */
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			C[i][j] = original[i][j] - matrix[i][j];
			C[i][j] = C[i][j] * C[i][j];
			V[j] += C[i][j];
		}
	}
	
	/* Sum. */
	for (j = 0; j < width; j++) {
		sum += V[j];
	}
	
	return sum / (float)(width*height);
}

int read_tiff(img_c **buffer, char tiff[]) {
	uint32 image_offset = 0;
	tsize_t strip_bytes;
	int strips, i, result;
	TIFF *image_in;
	
	/* Open the input file for reading. */
	if( (image_in = TIFFOpen(tiff, "r")) == NULL )
			return TIFF_OPEN_ERROR;
	
	/* Read some TIFF tags from the image. We need these for writing later. */
	TIFFGetField(image_in, TIFFTAG_IMAGEWIDTH, &width);
	TIFFGetField(image_in, TIFFTAG_IMAGELENGTH, &height);
	TIFFGetField(image_in, TIFFTAG_PHOTOMETRIC, &photo);
	TIFFGetField(image_in, TIFFTAG_BITSPERSAMPLE, &bpp);
	TIFFGetField(image_in, TIFFTAG_SAMPLESPERPIXEL, &spp);
	TIFFGetField(image_in, TIFFTAG_FILLORDER, &fillorder);
	TIFFGetField(image_in, TIFFTAG_COMPRESSION, &compression);
	TIFFGetField(image_in, TIFFTAG_ROWSPERSTRIP, &rows_per_strip);
	TIFFGetField(image_in, TIFFTAG_STRIPOFFSETS, &image_offset);
	TIFFGetField(image_in, TIFFTAG_PLANARCONFIG, &planar);
	TIFFGetField(image_in, TIFFTAG_ORIENTATION, &orientation);
	TIFFGetField(image_in, TIFFTAG_RESOLUTIONUNIT, &res_unit);
	TIFFGetField(image_in, TIFFTAG_XRESOLUTION, &xres);
	TIFFGetField(image_in, TIFFTAG_YRESOLUTION, &yres);
	strip_bytes = TIFFStripSize(image_in);
	strips = TIFFNumberOfStrips(image_in);
	
	/* Create a buffer to hold our image data and fill it with zeros. */
	bufferSize = strips * strip_bytes * sizeof(img_c);
	*buffer = _TIFFmalloc(bufferSize);
	if ( buffer == NULL )
		return TIFF_MEM_ERROR;
	_TIFFmemset(*buffer, 0x00, bufferSize);
	
	/* Read the image into our buffer. */
	for (i = 0; i < strips; i++) {
		result = TIFFReadRawStrip(image_in, i, *buffer + (i*strip_bytes),
								  strip_bytes);
		if ( result == -1 )
			return TIFF_READ_ERROR;
	}
	
	/* DEBUG: Write the raw data to a text file. */
	if ( DEBUG != 0 ) {
		result = write_debug(&(*buffer), "debug_read.txt");
		if ( result != 0 ) {
			print_error(result);
			return result;
		}
	}
	
	TIFFClose(image_in);
		
	printf("Input image specifications:\n");
	printf("%ix%i\t photo = %i\t %ibpp\t %ispp\t order = %i\n", width, height, 
			photo, bpp, spp, fillorder);
	printf("res_unit = %i\txres = %f\tyres = %f\n", res_unit, xres, yres);
	printf("compression: %i\n", compression);
	printf("rows per strip: %i\tstrips: %i\tbytes per strip: %i\n", 
			rows_per_strip, strips, strip_bytes);

	return 0;
}

int write_tiff(img_c **buffer, char tiff[]) {
	uint32 strip_size;
	int strips;
	TIFF *image_out;
	int y;
	
	/* DEBUG: Write the raw data to a text file. */
	if ( DEBUG != 0 ) {
		int result;
		result = write_debug(&(*buffer), "debug_write.txt");
		if ( result != 0 ) {
			print_error(result);
			return result;
		}
	}
	
	/* Fix some values that might be broken. */
	if ( fillorder == 0 )
		fillorder = FILLORDER_MSB2LSB;
	if ( spp == 0 )
		spp = 1;
	if ( orientation == 0 )
		orientation = ORIENTATION_TOPLEFT;
	
	/* Open the input file for writing. */
	if( (image_out = TIFFOpen(tiff, "w")) == NULL )
			return TIFF_WRITE_ERROR;
	
	/* Set some basic TIFF tags. */
	TIFFSetField(image_out, TIFFTAG_IMAGEWIDTH, width);
	TIFFSetField(image_out, TIFFTAG_IMAGELENGTH, height);
	TIFFSetField(image_out, TIFFTAG_PHOTOMETRIC, photo);
	TIFFSetField(image_out, TIFFTAG_BITSPERSAMPLE, bpp);
	TIFFSetField(image_out, TIFFTAG_SAMPLESPERPIXEL, spp);
	TIFFSetField(image_out, TIFFTAG_FILLORDER, fillorder);
	TIFFSetField(image_out, TIFFTAG_COMPRESSION, compression);
	TIFFSetField(image_out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);
	TIFFSetField(image_out, TIFFTAG_PLANARCONFIG, planar);
	TIFFSetField(image_out, TIFFTAG_ORIENTATION, orientation);
	TIFFSetField(image_out, TIFFTAG_RESOLUTIONUNIT, res_unit);
	TIFFSetField(image_out, TIFFTAG_XRESOLUTION, xres);
	TIFFSetField(image_out, TIFFTAG_YRESOLUTION, yres);
	
/*	strip_size = TIFFDefaultStripSize(image_out, 8000);
	strips = height / rows_per_strip; */
	
	/* Write the information to the file. */
	for (y = 0; y < height; y++) {
		TIFFWriteScanline(image_out, &(*buffer)[y*width], y, 0);
	}
	
	/* Close the file. */
	TIFFClose(image_out);
	
	return 0;
}