/*
	Author: James Sumners (james dot sumners AT gmail dot com)
	URL: http://james.roomfullofmirrors.com/
	License: Creative Commons 3.0 (NCSA)
		http://creativecommons.org/licenses/by-nc-sa/3.0/us/
	Attribution: Provide a name credit and link to URL above.
*/

#include "globals.h"

/* See 
* @article{MULC96,
*	ISSN = {0025-570X},
*	author = {Colm Mulcahy},
*	copyright = {Copyright 1996 Mathematical Association of America},
*	journal = {Mathematics Magazine},
*	month = {December},
*	number = {5},
*	pages = {323--343},
*	publisher = {Mathematical Association of America},
*	title = {Plotting and Scheming with Wavelets},
*	volume = {69},
*	year = {1996},
*}
*
* or
*
@article{MULC97,
*	author = {Colm Mulcahy},
*	journal = {Spelman College Science \& Mathematics Journal},
*	year = {1997},
*	month = {April},
*	volume = {1},
*	number = {1},
*	pages = {22--32},
*	title = {Image Compression Using The Haar Wavelet Transform},
*}
*
* for averaging and differencing process. */

/*
* Compresses the image using the sum and difference method.
* Takes the image buffer and a thresholding value as input.
*/
void haar_compress(img_f mat[][width], int threshold) {
	int i, j, k;
	float avgbuf[width/2]; /* Row vectors for temporary storage. */
	float difbuf[width/2];
	float SQRT = sqrtf(2.0);
	int x = width, y = height; /* Since width,height are global
								  variables, and we need to modify
								  them, we create some local variables x,y. */
	char fileout[32]; /* Used for debugging the loops. */
	
	/* Perform averaging and differencing until a single average entry is
	   present. */
	
	/* First we need to average and difference the rows. */
	printf("Transforming rows.\n");
	for (x; x > 1; x = x/2) {
		for (i = 0; i < height; i++) { /* rows */
			k = 0; /* k is the index for the avg/diff buffers */
			for (j = 0; j < x; j += 2) { /* cols */
				avgbuf[k] = (mat[i][j] + mat[i][j+1]) / SQRT;
				difbuf[k] = (mat[i][j] - mat[i][j+1]) / SQRT;
				k++;
			}
			
			/* Overwrite the row processed with the newly created row. */
			for (j = 0; j < x; j++) {
				mat[i][j] = ( j < (x/2) ) ? avgbuf[j] : difbuf[j-(x/2)];
			}
		}
	}
	
	/* Next we need to average and difference the columns. */
	printf("Transforming columns.\n");
	for (y; y > 1; y = y/2) {
		for (j = 0; j < width; j++) { /* cols */
			k = 0;
			for (i = 0; i < y; i += 2) { /* rows */
				avgbuf[k] = (mat[i][j] + mat[i+1][j]) / SQRT;
				difbuf[k] = (mat[i][j] - mat[i+1][j]) / SQRT;
				k++;
			}
			
			/* Overwrite the column processed with the newly created column. */
			for (i = 0; i < y; i++) {
				mat[i][j] = ( i < (y/2) ) ? avgbuf[i] : difbuf[i-(y/2)];
			}
		}
	}
	
	/* Finally, we need to apply thresholding if desired. */
	if (threshold != 256) { /* 256 = no more data */
		printf("\nApplying thresholding level %i.\n", threshold);
		for (i = 0; i < height; i++) {
			for (j = 0; j < width; j++) {
				mat[i][j] = (fabs(mat[i][j]) < threshold) ? 0 : mat[i][j];
			}
		}
	}
}

/*
* Decompresses the image using the sum and difference method.
*/
void haar_decompress(img_f mat[][width]) {
	int i, j, k;
	img_f mat_new[height][width];
	int x = width, y = height;
	float avg;
	float SQRT = sqrtf(2.0);
	int b;
	char fileout[64];
	
	/* Initialize the temporary matrix. */
	memcpy(mat_new, mat, width*height*sizeof(img_f));
	
	/* The variable 'b' keeps track of the width/height of the "averages" versus
	   the detail coefficients in the columns/rows. */
	printf("Restoring columns.\n");
	for (b = 1; b < height; b = b*2) {
		/* First we need to reverse the columns. */
		for (j = 0; j < width; j++) {
			k = 0;
			for (i = 0; k < b; i += 2) {
				avg = mat[k][j];
				/* Revert difference. */
				mat_new[i][j] = ((avg + mat[b+k][j])*SQRT) / 2.0;
				/* Revert averaging. */
				mat_new[i+1][j] = ((avg - mat[b+k][j])*SQRT) / 2.0;
				k++;
			}
		}
		/* Need to copy our work to the compressed matrix. */
		memcpy(mat, mat_new, width*height*sizeof(img_f));
	}
	
	/* Next we need to reverse the rows. */
	printf("Restoring rows.\n");
	for (b = 1; b < width; b = b*2) {
		for (i = 0; i < height; i++) {
			k = 0;
			for (j = 0; k < b; j += 2) {
				avg  = mat[i][k];
				/* Revert difference. */
				mat_new[i][j] = ((avg + mat[i][b+k])*SQRT) / 2.0;
				/* Revert averaging. */
				mat_new[i][j+1] = ((avg - mat_new[i][b+k])*SQRT) / 2.0;
				k++;
			}
		}
		
		memcpy(mat, mat_new, width*height*sizeof(img_f));
	}
}